﻿#region Copyright
/// <summary>The ScummVM WPF UI project</summary>
/// <copyright>(c) 2016 MojoTouch</copyright>
/// <author>Ilya Rovensky</author>
#endregion

using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using System.Windows;
using System.Windows.Input;

namespace WpfInterface
{
    /// <summary>
    /// Exposed methods for interaction between the WPF subsystem and the game engine
    /// </summary>
    public class UserInterface
    {
        private UIWindow _uiWindow;
        private List<string> _strings;
        private List<Setting> _settings;
        private PopUp _currentPopUp;
#if TARGET_T7G
        private bool _isQuitInitiated = false;
        private bool _splashWasShown = false;
#endif
        private GameType _gameToPromote;
        private string _urlToPromote;
        private string _urlToRate;

        /// <summary>
        /// Initializes the UI thread, creates initial windows, starts the WPF message queue
        /// </summary>
        public UserInterface()
        {
            Log("Starting the WPF subsystem...");

            using (ManualResetEventSlim waitEvent = new ManualResetEventSlim())
            {
                // create main ui thread for the WPF subsystem
                Thread uiThread = new Thread(() =>
                {
                    _uiWindow = new UIWindow();
                    waitEvent.Set();
                    new App().Run(_uiWindow);
                });

                uiThread.SetApartmentState(ApartmentState.STA);
                uiThread.Start();

                //return from the constructor after the application is up and running
                waitEvent.Wait();
            }

            Log("The WPF subsystem started.");
        }

        /// <summary>
        /// Logs a message to the console
        /// </summary>
        /// <param name="message">A message to display, an exception or any other object</param>
        public static void Log(object message)
        {
#if DEBUG
            Console.WriteLine(string.Format("************ {0: yyyy/MM/dd hh.mm.ss.fff}: [WPF UI] {1}", DateTime.Now, message));
#else
            // causes hanging up in the middle of the game
            //File.AppendAllText("log.txt", string.Format("{0: yyyy/MM/dd hh.mm.ss.fff}: {1}\n", DateTime.Now, message));
#endif
        }

        /// <summary>
        /// Executes an action on the UI thread and waits for its completion
        /// </summary>
        /// <param name="action">An action to run</param>
        /// <param name="methodDescription">Description of the method</param>
        private void ProtectedExecute(Action action, string methodDescription)
        {

            Log(string.Format("Entering {0}.", methodDescription));

            if (Application.Current != null)
                try
                {
                    Application.Current.Dispatcher.Invoke(action);
                }
                catch (Exception e)
                {
                    Log(string.Format("Error in {0}: {1}", methodDescription, e));
                }
            else
                MessageBox.Show($"{nameof(ProtectedExecute)}: The UI thread is already destroyed.");

            Log(string.Format("Leaving {0}.", methodDescription));
        }

        /// <summary>
        /// Shows the main menu and the splash screen, if needed
        /// </summary>
        /// <param name="menuAction">Menu to show</param>
        /// <returns>The selected action from the main menu</returns>
        public MenuAction ShowMenu(MenuAction menuAction, int menuState = 0)
        {
            // workaround for the bug with incorrect quitting
            //if (menuAction == MenuAction.Splash && _splashWasShown) // this means we are closing
            //    return MenuAction.Quit;

            Log(string.Format("MenuAction = {0}, menuState = {1}.", menuAction, menuState));

            ProtectedExecute(() =>
            {
                switch (menuAction)
                {
                    case MenuAction.Splash:
                        _uiWindow.CurrentMenu = new SplashScreen();
                        break;
                    case MenuAction.Main:
                        _uiWindow.CurrentMenu = new MainMenu(menuState, _gameToPromote, _urlToPromote) { Strings = _strings };
#if TARGET_T7G
                        if (!_splashWasShown)
                        {
                            MenuState State = new MenuState(menuState);
                            if (State.IsLoadEnabled || State.IsContinueEnabled)
                            {
                                System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.CurrentUICulture;
                                Log("Playing WAV: back_for_more");
                                switch (ci.TwoLetterISOLanguageName)
                                {
                                    case "en":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/en/back_for_more.wav").Play();
                                        break;
                                    case "de":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/de/back_for_more.wav").Play();
                                        break;
                                    case "fr":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/fr/back_for_more.wav").Play();
                                        break;
                                    case "ru":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/ru/back_for_more.wav").Play();
                                        break;
                                }
                            }
                            else
                            {
                                System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.CurrentUICulture;
                                Log("Playing WAV: welcome_to_my_house");
                                switch (ci.TwoLetterISOLanguageName)
                                {
                                    case "en":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/en/welcome_to_my_house.wav").Play();
                                        break;
                                    case "de":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/de/welcome_to_my_house.wav").Play();
                                        break;
                                    case "fr":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/fr/welcome_to_my_house.wav").Play();
                                        break;
                                    case "ru":
                                        new System.Media.SoundPlayer($"{Assets.Path}/wav/ru/welcome_to_my_house.wav").Play();
                                        break;
                                }
                            }
                            _splashWasShown = true;
                        }
#endif
                        // start music
                        //SoundPlayer.Music(true);
                        break;
                    case MenuAction.Settings:
                        _uiWindow.CurrentMenu = new SettingsMenu() { Strings = _strings };
                        break;
                    case MenuAction.SettingsAbout:
                        _uiWindow.CurrentMenu = new AboutSubMenu(_urlToRate) { Strings = _strings };
                        break;
                    case MenuAction.Tutorial:
                        _uiWindow.CurrentMenu = new VideoScreen();
                        break;
                    case MenuAction.Extras:
                        _uiWindow.CurrentMenu = new ExtrasMenu() { Strings = _strings };
                        break;
                    case MenuAction.SettingsVoice:
                    case MenuAction.SettingsControls:
                    case MenuAction.SettingsGraphics:
                    case MenuAction.SettingsLanguage:
                    case MenuAction.SettingsMusic:
                        if (_uiWindow.CurrentMenu is SettingsSubMenu)
                            ((SettingsSubMenu)_uiWindow.CurrentMenu).ShowAgain(_settings);                        
                        else
                            _uiWindow.CurrentMenu = new SettingsSubMenu(_settings) { Strings = _strings };
                        break;
                    case MenuAction.DeveloperLogo1:
                    case MenuAction.DeveloperLogo2:
                        _uiWindow.CurrentMenu = new LogoScreen(menuAction == MenuAction.DeveloperLogo2);
                        break;
                }
            }, nameof(ShowMenu));

            menuAction = _uiWindow.CurrentMenu.Result;
#if TARGET_T7G
            if (menuAction == MenuAction.Quit)
            {
                _isQuitInitiated = true;
                Log("_isQuitInitiated = true");
            }
            else
            {
                _isQuitInitiated = false;
            }
#endif
            /*
            if (menuAction == MenuAction.Splash)
                if (!_isQuitInitiated)
                {
                    Log("Show the main menu after the splash screen.");
                    return ShowMenu(MenuAction.Main, menuState);
                }
                else
                    menuAction = MenuAction.Quit;
                    */

            Log(string.Format("Result = {0}.", menuAction));

            /*
            if (menuAction == MenuAction.Extras ||
                menuAction == MenuAction.OpenHouse ||
                menuAction == MenuAction.Tutorial)
                // stop music
                ProtectedExecute(() => SoundPlayer.Music(false), nameof(SoundPlayer));
            */
            return menuAction;
        }

        public int ShowSettings(MenuAction menuAction, List<Setting> settings)
        {
            Log("Entering ShowSettings");
            _settings = settings;           
            return ShowMenu(menuAction, 0) == MenuAction.Settings ? 0 : 1;
        }

        public void HideMenu()
        {
            ProtectedExecute(() => _uiWindow.CurrentMenu = null, nameof(HideMenu));
        }

        private int ShowDialog(Func<PopUp> popUpFactory)
        {
            ProtectedExecute(() =>
            {
                //_uiWindow.Activate();
                _currentPopUp = popUpFactory();
                //_currentPopUp.Loaded += (_, __) => { _uiWindow.Activate(); _currentPopUp.Activate(); };
                _currentPopUp.ShowDialog();
            }, nameof(ShowDialog));

            int result = (int)_currentPopUp.Result;
            _currentPopUp = null;

            return result;
        }

        public int ShowAlertDialog(string message, List<string> buttons)
        {
            AlertPopUp.AlertData alertData = new AlertPopUp.AlertData
            {
                Message = message,
                Buttons = buttons
            };

            int result = ShowDialog(() => new AlertPopUp(alertData));

            Log(string.Format("ShowAlertDialog, Result {0}.", result));

#if TARGET_T7G
            if (message.Contains("Exit"))
            {
                _isQuitInitiated = true;
            }

            if (_isQuitInitiated && result == 31)
            {
                System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.CurrentUICulture;
                Log("Playing WAV: come_back");
                switch (ci.TwoLetterISOLanguageName)
                {
                    case "en":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/en/come_back.wav").PlaySync();
                        break;
                    case "de":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/de/come_back.wav").Play();
                        break;
                    case "fr":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/fr/come_back.wav").Play();
                        break;
                    case "ru":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/ru/come_back.wav").Play();
                        break;
                }
                _isQuitInitiated = false;
            }
#endif
            return result;
        }

        public void ShowProgressDialog(string message)
        {
            ProtectedExecute(() =>
            {
                ProgressPage progressPage = new ProgressPage();
                progressPage.MessageLabel.Text = message;
                _currentPopUp = progressPage;
                progressPage.Show();
                progressPage.Activate();
            }, nameof(ShowProgressDialog));
            Thread.Sleep(1000);
        }

        public void HideProgressDialog()
        {
            if (_currentPopUp != null)
            {
                ProtectedExecute(() => _currentPopUp.Close(), nameof(HideProgressDialog));
                _currentPopUp = null;
            }
            //HideMenu();
        }

        public int ShowDialogWithList(string message, List<string> items, int buttonNum, string negativeButton, string positiveButton)
        {
            ListPopUp.ListData listData = new ListPopUp.ListData
            {
                Message = message,
                Slots = items,
                ButtonCount = buttonNum,
                NegativeButtonTitle = negativeButton,
                PositiveButtonTitle = positiveButton
            };

            int result = ShowDialog(() => new ListPopUp(listData));

            return result;
        }

        public int ShowDialogWithInput(string message, ref string input, string negativeButton, string positiveButton)
        {
            InputPopUp.InputData inputData = new InputPopUp.InputData
            {
                Message = message,
                InitialInput = input,
                NegativeButtonTitle = negativeButton,
                PositiveButtonTitle = positiveButton
            };

            InputPopUp inputPopup = null;

            int result = ShowDialog(() =>
            {
                inputPopup = new InputPopUp(inputData);
                return inputPopup;
            });

            string text = null;
            ProtectedExecute(() => text = inputPopup.MessageTextBox.Text, nameof(ShowDialogWithInput));
            input = string.IsNullOrWhiteSpace(text) ? "No name" : text;

            Log(string.Format("ShowDialogWithInput, Result {0}.", result));

#if TARGET_T7G

            if (result == 28)
            { 
                System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.CurrentUICulture;
                Log("Playing WAV: game_has_been_saved");
                switch (ci.TwoLetterISOLanguageName)
                {
                    case "en":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/en/game_has_been_saved.wav").Play();
                        break;
                    case "de":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/de/game_has_been_saved.wav").Play();
                        break;
                    case "fr":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/fr/game_has_been_saved.wav").Play();
                        break;
                    case "ru":
                        new System.Media.SoundPlayer($"{Assets.Path}/wav/ru/game_has_been_saved.wav").Play();
                        break;
                }
            }
#endif
            return result;
        }

        public void ShowNotification(string message)
        {
            //ProtectedExecute(() =>
            //{
            //    using (ManualResetEvent waitEvent = new ManualResetEvent(false))
            //    {
            //        ExecuteOnUIThread(() =>
            //        {
            //            _currentMenu = new NotificationPopUp(message);

            //            PropertyChangedEventHandler handler = null;
            //            handler = (s, e) =>
            //            {
            //                if (e.PropertyName == nameof(_currentMenu.Result))
            //                {
            //                    _currentMenu.PropertyChanged -= handler;
            //                    waitEvent.Set();
            //                }
            //            };
            //            _currentMenu.PropertyChanged += handler;

            //            //_uiWindow.ActivateContent(true);
            //            _uiWindow.Navigate(_currentMenu);
            //        });

            //        waitEvent.WaitOne();
            //        //_uiWindow.ActivateContent(false);
            //    }
            //}, "ShowNotification");
        }

        // http://stackoverflow.com/questions/5673556/is-it-possible-to-let-my-c-sharp-wpf-program-know-if-the-user-has-a-touchscreen
        public bool HasTouchInput()
        {
            bool result = false;

            ProtectedExecute(() =>
            {
                foreach (TabletDevice tabletDevice in Tablet.TabletDevices)
                    if (tabletDevice.Type == TabletDeviceType.Touch)
                    {
                        result = true;
                        break;
                    }
            }, nameof(HasTouchInput));

            return result;
        }

        public void InitStrings(List<string> strings, int gameToPromote, string urlToPromote, string urlToRate)
        {
            _strings = strings;
            _gameToPromote = (GameType)gameToPromote;
            _urlToPromote = urlToPromote;
            _urlToRate = urlToRate;
        }

        public bool IsShowingMenu()
        {
            bool result = _uiWindow.CurrentMenu != null;
            Log($"Showing menu: {result}");
            return result;
        }

        public void SetOglWindowHandle(IntPtr handle)
        {
            //if (OglWindowHandle == IntPtr.Zero && handle != IntPtr.Zero)
            //    Win32Window.EnableAltTab(_uiWindow, false);
            OglWindowHandle = handle;
        }

        public static IntPtr OglWindowHandle { get; set; } = IntPtr.Zero;

        public void DisplayMainMenu(int menuState)
        {
            ProtectedExecute(() => _uiWindow.CurrentMenu = 
            new MainMenu(menuState, _gameToPromote, _urlToPromote) { Strings = _strings }, nameof(DisplayMainMenu));
        }

        public string QueryUrl(string url)
        {
            string result = "stream was empty.";
            try
            {
                using (WebClient webClient = new WebClient())
                    result = webClient.DownloadString(url);
            }
            catch
            {
                // suppress possible problems
                return result;
            }
            return result;
        }
    }
}
